# log21.FileHandler.py
# CodeWriter21

from logging import FileHandler as _FileHandler
from log21.Formatters import DecolorizingFormatter as _DecolorizingFormatter


class DecolorizingFileHandler(_FileHandler):
    terminator = ''

    def emit(self, record):
        if self.stream is None:
            self.stream = self._open()
        try:
            msg = self.format(record)
            msg = _DecolorizingFormatter.decolorize(msg)
            stream = self.stream
            stream.write(msg + self.terminator)
            self.flush()
        except Exception:
            self.handleError(record)



# log21.TreePrint.py
# CodeWriter21

from typing import Union as _Union

from log21.Colors import get_colors as _gc


class TreePrint:
    class Node:
        colors = {
            'branches': _gc('Green'),
            'fruit': _gc('LightMagenta'),
        }

        def __init__(self, value: _Union[str, int], children: list = None, indent: int = 4, colors: dict = None,
                     mode='-'):
            self.value = str(value)
            if children:
                self._children = children
            else:
                self._children = []
            self.indent = indent
            self.colors = self.colors.copy()
            if colors:
                for key, value in colors.items():
                    if key in self.colors:
                        self.colors[key] = _gc(value)
            if not mode:
                self.mode = 1
            else:
                self.mode = self._get_mode(mode)
                if self.mode == -1:
                    raise ValueError('`mode` must be - or =')

        def _get_mode(self, mode=None) -> int:
            if not mode:
                mode = self.mode
            if isinstance(mode, int):
                if mode in [1, 2]:
                    return mode
            if mode in '-_─┌│|└┬├└':
                return 1
            if mode in '=═╔║╠╚╦╚':
                return 2
            return -1

        def __str__(self, level=0, prefix='', mode=None):
            mode = self._get_mode(mode)
            if mode == -1:
                raise ValueError('`mode` must be - or =')
            chars = '─┌│└┬├└'
            if mode == 2:
                chars = '═╔║╚╦╠╚'

            text = _gc(self.colors['branches']) + prefix
            if level == 0:
                text += chars[0]  # ─ OR ═
                prefix += ' '

            if self.has_child():
                text += chars[4]  # ┬ OR ╦
            else:
                text += chars[0]  # ─ OR ═

            text += ' ' + _gc(self.colors['fruit']) + str(self.value) + '\n'

            for i, child in enumerate(self._children):
                prefix_ = ''
                for j in range(len(prefix)):
                    if prefix[j] in '┌│├┬╔║╠╦':
                        prefix_ += chars[2]  # │ OR ║
                    elif prefix[j] in chars:
                        prefix_ += ' '
                    else:
                        prefix_ += prefix[j]

                if i + 1 == len(self._children):
                    prefix_ += chars[6]  # └ OR ╚
                else:
                    prefix_ += chars[5]  # ├ OR ╠
                prefix_ += chars[0] * (self.indent - 1)  # ─ OR ═
                prefix_ = prefix_[:len(prefix)] + _gc(self.colors['branches']) + prefix_[len(prefix):]
                text += child.__str__(level=level + 1, prefix=prefix_, mode=mode)

            return text

        def has_child(self):
            return len(self._children) > 0

        def add_child(self, child: 'TreePrint.Node'):
            if not isinstance(child, TreePrint.Node):
                raise TypeError('`child` must be TreePrint.Node')
            self._children.append(child)

        def get_child(self, value: str = None, index: int = None):
            if value and index:
                raise ValueError('`value` and `index` can not be both set')
            if not value and not index:
                raise ValueError('`value` or `index` must be set')
            if value:
                for child in self._children:
                    if child.value == value:
                        return child
            if index:
                return self._children[index]

        @staticmethod
        def add_to(node: 'TreePrint.Node', data: _Union[dict, list, str, int], indent: int = 4, colors: dict = None,
                   mode='-'):
            if isinstance(data, dict):
                if len(data) == 1:
                    child = TreePrint.Node(list(data.keys())[0], indent=indent, colors=colors, mode=mode)
                    TreePrint.Node.add_to(child, list(data.values())[0], indent=indent, colors=colors, mode=mode)
                    node.add_child(child)
                else:
                    for key, value in data.items():
                        child = TreePrint.Node(key, indent=indent, colors=colors, mode=mode)
                        TreePrint.Node.add_to(child, value, indent=indent, colors=colors, mode=mode)
                        node.add_child(child)
            elif isinstance(data, list):
                for value in data:
                    if isinstance(value, dict):
                        for key, dict_value in value.items():
                            child = TreePrint.Node(key, indent=indent, colors=colors, mode=mode)
                            TreePrint.Node.add_to(child, dict_value, indent=indent, colors=colors, mode=mode)
                            node.add_child(child)
                    elif isinstance(value, list):
                        child = TreePrint.Node('>', indent=indent, colors=colors, mode=mode)
                        TreePrint.Node.add_to(child, value, indent=indent, colors=colors, mode=mode)
                        node.add_child(child)
                    else:
                        child = TreePrint.Node(str(value), indent=indent, colors=colors, mode=mode)
                        node.add_child(child)
            else:
                child = TreePrint.Node(str(data), indent=indent, colors=colors, mode=mode)
                node.add_child(child)

    def __init__(self, data: _Union[dict, list, str, int], indent: int = 4, colors: dict = None, mode='-'):
        self.indent = indent
        self.mode = mode
        if isinstance(data, dict):
            if len(data) == 1:
                self.root = self.Node(list(data.keys())[0], indent=indent, colors=colors)
                self.add_to_root(list(data.values()), colors=colors)
            else:
                self.root = self.Node('root', indent=indent, colors=colors)
                self.add_to_root(data, colors=colors)
        elif isinstance(data, list):
            self.root = self.Node('root', indent=indent, colors=colors)
            self.add_to_root(data, colors=colors)
        else:
            self.root = self.Node(str(data), indent=indent, colors=colors)

    def add_to_root(self, data: _Union[dict, list, str, int], colors: dict = None):
        self.Node.add_to(self.root, data, indent=self.indent, colors=colors, mode=self.mode)

    def __str__(self, mode=None):
        if not mode:
            mode = self.mode
        return self.root.__str__(mode=mode)


def tree_format(data: _Union[dict, list, str, int], indent: int = 4, mode='-', colors: dict = None):
    return str(TreePrint(data, indent=indent, colors=colors, mode=mode))



# log21.Levels.py
# CodeWriter21

import logging as _logging

__all__ = ['CRITICAL', 'FATAL', 'ERROR', 'WARNING', 'WARN', 'INFO', 'DEBUG', 'NOTSET']

CRITICAL = _logging.CRITICAL
FATAL = CRITICAL
ERROR = _logging.ERROR
WARNING = _logging.WARNING
WARN = WARNING
INFO = _logging.INFO
DEBUG = _logging.DEBUG
NOTSET = _logging.NOTSET



# log21.__init__.py
# CodeWriter21

import io as _io
import logging as _logging

from typing import Union as _Union, Dict as _Dict

from log21.Levels import *
from log21.Logger import Logger
from log21.Manager import Manager
from log21.ProgressBar import ProgressBar
from log21.PPrint import PrettyPrinter, pformat
from log21.TreePrint import TreePrint, tree_format
from log21.Argparse import ColorizingArgumentParser
from log21.FileHandler import DecolorizingFileHandler
from log21.StreamHandler import ColorizingStreamHandler, StreamHandler
from log21.Formatters import ColorizingFormatter, DecolorizingFormatter
from log21.Colors import Colors, get_color, get_colors, ansi_esc, get_color_name, closest_color

__version__ = "1.5.10"
__author__ = "CodeWriter21 (Mehrad Pooryoussof)"
__github__ = "Https://GitHub.com/MPCodeWriter21/log21"
__all__ = ['ColorizingStreamHandler', 'DecolorizingFileHandler', 'ColorizingFormatter', 'DecolorizingFormatter',
           'get_logger', 'Logger', 'Colors', 'get_color', 'get_colors', 'CRITICAL', 'FATAL', 'ERROR', 'WARNING', 'WARN',
           'INFO', 'DEBUG', 'NOTSET', 'StreamHandler', 'ColorizingArgumentParser', 'PrettyPrinter', 'pformat',
           'pprint', 'pretty_print', 'tree_format', 'TreePrint', 'Manager', 'get_color_name', 'closest_color',
           'ansi_esc', '__version__', '__author__', '__github__', 'debug', 'info', 'warning', 'warn', 'error',
           'critical', 'fatal', 'exception', 'log', 'basic_config', 'basicConfig', 'ProgressBar', 'progress_bar']

_manager = Manager()
_logging.setLoggerClass(Logger)


def get_logger(name: str = '', level: _Union[int, str] = NOTSET, show_time: bool = True,
               show_level: bool = True, colorize_time_and_level: bool = True, fmt: str = None,
               datefmt: str = "%H:%M:%S", style: str = '%', handle_carriage_return: bool = True,
               handle_new_line: bool = True, override=False, level_names: _Dict[int, str] = None) -> Logger:
    """
    Returns a logging.Logger with colorizing support.
    >>>
    >>> import log21
    >>>
    >>> l = log21.get_logger()
    >>> l.warning('Pretty basic, huh?')
    [14:49:41] [WARNING] Pretty basic, huh?
    >>> l.critical('CONTINUE READING!! please...')
    [14:50:08] [CRITICAL] CONTINUE READING!! please...
    >>>
    >>> my_logger = log21.get_logger(name='CodeWriter21', level=log21.INFO, fmt='{asctime} -> [{levelname}]: {message}',
    ... style='{', override=True)
    >>>
    >>> my_logger.info('FYI: My name is Mehrad.')
    14:56:12 -> [INFO]: FYI: My name is Mehrad.
    >>> my_logger.error(log21.get_color('LightRed') + 'Oh no! Something went wrong D:')
    14:56:29 -> [ERROR]: Oh no! Something went wrong D:
    >>>
    >>> my_logger.debug(1 ,2 ,3)
    >>> # It prints Nothing because our logger level is INFO and DEBUG level is less than INFO.
    >>> # So let's modify the my_logger's level
    >>> my_logger.setLevel(log21.DEBUG)
    >>> # Now we try again...
    >>> my_logger.debug(1, 2, 3)
    14:57:34 -> [DEBUG]: 1 2 3
    >>> # Well Done. Right?
    >>> # Let's see more
    >>> my_logger.debug('I like %s number!', args=('21', ), end='\033[0m\n\n\n')
    15:01:43 -> [DEBUG]: I like 21 number!


    >>> # Well, I've got a question...
    >>> # Do you know the name of this color?
    >>> # #888888
    >>> # Oh ya! I can use get_color_name
    >>> log21.get_color_name('#888888')
    'gray'
    >>> # Oh thank you dear!
    >>> # Yes I knew that was grey -_- But I wanted to introduce my little friend ☺
    >>> # See you soon!
    >>>

    :param name: Optional[str]: The name of the logger
    :param level: Union[int, str] = logging.NOTSET: The logging level of the logger
    :param show_time: bool = True: Show the time in the log
    :param show_level: bool = True: Show the level of logging in the log
    :param fmt: Optional[str]: Custom formatting for the logger - overrides the default(show_time & show_level)
    :param datefmt: str = "%H:%M:%S": Custom date-time formatting for the logger
    :param style: str = '%': Use a style parameter of '%', '{' or '$' to specify that you want to use one of %-formatting,
        :meth:`str.format` (``{}``) formatting or :class:`string.Template` formatting in your format string.
    :param colorize_time_and_level: bool = True: Colorizes the time and level using the default colors
    :param handle_carriage_return: bool = True: Adds a line of space characters to remove any text before the CR
    :param handle_new_line: bool = True: Places the NewLine characters at the beginning of the text before everything else
    :param override: bool = True: Overrides the logger attributes even if it already exists
    :param level_names: Dict[int, str] = None: You can specify custom level names.
    :return: logging.Logger

    """
    if not isinstance(name, str):
        raise TypeError('A logger name must be a string')
    logger = None
    if name:
        logger = _manager.getLogger(name)
    if (not logger) or override:
        logger = Logger(name, level)
        # Prepares a formatting if the fmt was None
        if not fmt:
            fmt = "%(message)s"
            if show_level:
                fmt = "[%(levelname)s] " + fmt
            if show_time:
                fmt = "[%(asctime)s] " + fmt
            fmt = '\r' + fmt
        # Defines the formatter
        formatter = ColorizingFormatter(fmt, datefmt, style=style, level_names=level_names)
        if not colorize_time_and_level:
            for key in formatter.level_colors:
                formatter.level_colors[key] = tuple()
            formatter.time_color = tuple()

        # Defines the handler
        handler = ColorizingStreamHandler(handle_carriage_return=handle_carriage_return,
                                          handle_new_line=handle_new_line)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        _manager.addLogger(name, logger)
    return logger


getLogger = get_logger


def print(*msg, args: tuple = (), end='\033[0m\n', **kwargs):
    logger = get_logger('log21.print', level=DEBUG, show_time=False, show_level=False)
    logger.print(*msg, args=args, end=end, **kwargs)


def pprint(obj, indent=1, width=80, depth=None, signs_colors: _Dict[str, str] = None, *, compact=False, sort_dicts=True,
           underscore_numbers=False, end='\033[0m\n', **kwargs):
    logger = get_logger('log21.pprint', level=DEBUG, show_time=False, show_level=False)
    logger.print(pformat(obj=obj, indent=indent, width=width, depth=depth, signs_colors=signs_colors, compact=compact,
                         sort_dicts=sort_dicts, underscore_numbers=underscore_numbers), end=end, **kwargs)


pretty_print = pprint


def tree_print(obj, indent: int = 4, mode='-', colors: _Dict[str, str] = None, end='\033[0m\n', **kwargs):
    logger = get_logger('log21.tree_print', level=DEBUG, show_time=False, show_level=False)
    logger.print(tree_format(obj, indent=indent, mode=mode, colors=colors), end=end, **kwargs)


tprint = tree_print

root = Logger('root-logger', INFO)


def basic_config(**kwargs):
    """
    Do basic configuration for the logging system.

    This function does nothing if the root logger already has handlers
    configured, unless the keyword argument *force* is set to ``True``.
    It is a convenience method intended for use by simple scripts
    to do one-shot configuration of the logging package.

    The default behaviour is to create a ColorizingStreamHandler which writes to
    sys.stderr, set a formatter using the BASIC_FORMAT format string, and
    add the handler to the root logger.

    A number of optional keyword arguments may be specified, which can alter
    the default behaviour.

    filename  Specifies that a FileHandler be created, using the specified
              filename, rather than a StreamHandler.
    filemode  Specifies the mode to open the file, if filename is specified
              (if filemode is unspecified, it defaults to 'a').
    format    Use the specified format string for the handler.
    datefmt   Use the specified date/time format.
    style     If a format string is specified, use this to specify the
              type of format string (possible values '%', '{', '$', for
              %-formatting, :meth:`str.format` and :class:`string.Template`
              - defaults to '%').
    level     Set the root logger level to the specified level.
    stream    Use the specified stream to initialize the StreamHandler. Note
              that this argument is incompatible with 'filename' - if both
              are present, 'stream' is ignored.
    handlers  If specified, this should be an iterable of already created
              handlers, which will be added to the root handler. Any handler
              in the list which does not have a formatter assigned will be
              assigned the formatter created in this function.
    force     If this keyword  is specified as true, any existing handlers
              attached to the root logger are removed and closed, before
              carrying out the configuration as specified by the other
              arguments.
    encoding  If specified together with a filename, this encoding is passed to
              the created FileHandler, causing it to be used when the file is
              opened.
    errors    If specified together with a filename, this value is passed to the
              created FileHandler, causing it to be used when the file is
              opened in text mode. If not specified, the default value is
              `backslashreplace`.

    Note that you could specify a stream created using open(filename, mode)
    rather than passing the filename and mode in. However, it should be
    remembered that StreamHandler does not close its stream (since it may be
    using sys.stdout or sys.stderr), whereas FileHandler closes its stream
    when the handler is closed.
    """
    force = kwargs.pop('force', False)
    encoding = kwargs.pop('encoding', None)
    errors = kwargs.pop('errors', 'backslashreplace')
    if force:
        for handler in root.handlers[:]:
            root.removeHandler(handler)
            handler.close()
    if len(root.handlers) == 0:
        handlers = kwargs.pop("handlers", None)
        if handlers is None:
            if "stream" in kwargs and "filename" in kwargs:
                raise ValueError("'stream' and 'filename' should not be specified together")
        else:
            if "stream" in kwargs or "filename" in kwargs:
                raise ValueError("'stream' or 'filename' should not be specified together with 'handlers'")
        if handlers is None:
            filename = kwargs.pop("filename", None)
            mode = kwargs.pop("filemode", 'a')
            if filename:
                if 'b' in mode:
                    errors = None
                else:
                    encoding = _io.text_encoding(encoding)
                handler = DecolorizingFileHandler(filename, mode, encoding=encoding, errors=errors)
            else:
                stream = kwargs.pop("stream", None)
                handler = ColorizingStreamHandler(stream=stream)
            handlers = [handler]
        date_format = kwargs.pop("datefmt", "%H:%M:%S")
        style = kwargs.pop("style", '%')
        if style not in '%{$':
            raise ValueError('Style must be one of: %, {, $')
        format_ = kwargs.pop("format", {'%': '[%(asctime)s] [%(levelname)s] %(message)s',
                                        '{': '[{asctime}] [{levelname}] {message}',
                                        '$': '[${asctime}] [${levelname}] ${message}'}[style])
        formatter = ColorizingFormatter(format_, date_format, style=style)
        for handler in handlers:
            if handler.formatter is None:
                handler.setFormatter(formatter)
            root.addHandler(handler)
        level = kwargs.pop("level", None)
        if level is not None:
            root.setLevel(level)
        if kwargs:
            keys = ', '.join(kwargs.keys())
            raise ValueError('Unrecognised argument(s): %s' % keys)


basicConfig = basic_config


def critical(*msg, args=(), **kwargs):
    """
    Log a message with severity 'CRITICAL' on the root logger. If the logger has no handlers, call basicConfig() to add
    a console handler with a pre-defined format.
    """
    if len(root.handlers) == 0:
        basic_config()
    root.critical(*msg, args=args, **kwargs)


def fatal(*msg, args=(), **kwargs):
    """
    Don't use this function, use critical() instead.
    """
    critical(*msg, args=args, **kwargs)


def error(*msg, args=(), **kwargs):
    """
    Log a message with severity 'ERROR' on the root logger. If the logger has no handlers, call basicConfig() to add a
    console handler with a pre-defined format.
    """
    if len(root.handlers) == 0:
        basic_config()
    root.error(*msg, args=args, **kwargs)


def exception(*msg, args=(), exc_info=True, **kwargs):
    """
    Log a message with severity 'ERROR' on the root logger, with exception information. If the logger has no handlers,
    basicConfig() is called to add a console handler with a pre-defined format.
    """
    error(*msg, args=args, exc_info=exc_info, **kwargs)


def warning(*msg, args=(), **kwargs):
    """
    Log a message with severity 'WARNING' on the root logger. If the logger has no handlers, call basicConfig() to add
    a console handler with a pre-defined format.
    """
    if len(root.handlers) == 0:
        basic_config()
    root.warning(*msg, args=args, **kwargs)


def warn(*msg, args=(), **kwargs):
    warning(*msg, args=args, **kwargs)


def info(*msg, args=(), **kwargs):
    """
    Log a message with severity 'INFO' on the root logger. If the logger has no handlers, call basicConfig() to add a
    console handler with a pre-defined format.
    """
    if len(root.handlers) == 0:
        basic_config()
    root.info(*msg, args=args, **kwargs)


def debug(*msg, args=(), **kwargs):
    """
    Log a message with severity 'DEBUG' on the root logger. If the logger has no handlers, call basicConfig() to add a
    console handler with a pre-defined format.
    """
    if len(root.handlers) == 0:
        basic_config()
    root.debug(*msg, args=args, **kwargs)


def log(level, *msg, args=(), **kwargs):
    """
    Log 'msg % args' with the integer severity 'level' on the root logger. If the logger has no handlers, call
    basicConfig() to add a console handler with a pre-defined format.
    """
    if len(root.handlers) == 0:
        basic_config()
    root.log(level, *msg, args=args, **kwargs)


def progress_bar(progress: float, total: float, width: int = None, prefix: str = '|', suffix: str = '|',
                 show_percentage: bool = True):
    """
    Print a progress bar to the console.
    """

    bar = ProgressBar(width=width, prefix=prefix, suffix=suffix, show_percentage=show_percentage)

    print(bar.get_bar(progress, total))



# log21.Logger.py
# CodeWriter21

import log21 as _log21
import logging as _logging
from logging import raiseExceptions as _raiseExceptions
from log21.Levels import *

__all__ = ['Logger']


class Logger(_logging.Logger):
    def __init__(self, name, level=NOTSET):
        super().__init__(name, level)
        self.setLevel(level)
        self._progress_bar = None

    def isEnabledFor(self, level):
        """
        Is this logger enabled for level 'level'?
        """

        return self.level <= level

    def log(self, level: int, *msg, args: tuple = (), end='\033[0m\n', **kwargs):
        """
        Log 'msg % args' with the integer severity 'level'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.log(level, "We have a %s", args=("mysterious problem",), exc_info=1)
        """
        msg = ' '.join([str(m) for m in msg]) + end
        if not isinstance(level, int):
            if _raiseExceptions:
                raise TypeError("level must be an integer")
            else:
                return
        if self.isEnabledFor(level):
            self._log(level, msg, args, **kwargs)

    def debug(self, *msg, args: tuple = (), end='\033[0m\n', **kwargs):
        """
        Log 'msg % args' with severity 'DEBUG'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.debug("Houston, we have a %s", args=("thorny problem",), exc_info=1)
        """
        if self.isEnabledFor(DEBUG):
            msg = ' '.join([str(m) for m in msg]) + end
            self._log(DEBUG, msg, args, **kwargs)

    def info(self, *msg, args: tuple = (), end='\033[0m\n', **kwargs):
        """
        Log 'msg % args' with severity 'INFO'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.info("Houston, we have an %s", args=("interesting problem",), exc_info=1)
        """
        if self.isEnabledFor(INFO):
            msg = ' '.join([str(m) for m in msg]) + end
            self._log(INFO, msg, args, **kwargs)

    def warning(self, *msg, args: tuple = (), end='\033[0m\n', **kwargs):
        """
        Log 'msg % args' with severity 'WARNING'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.warning("Houston, we have a %s", args=("bit of a problem",), exc_info=1)
        """
        if self.isEnabledFor(WARNING):
            msg = ' '.join([str(m) for m in msg]) + end
            self._log(WARNING, msg, args, **kwargs)

    warn = warning

    def write(self, *msg, args: tuple = (), end='', **kwargs):
        """
        Log 'msg % args' with severity 'WARNING'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.write("Houston, we have a %s", args=("bit of a problem",), exc_info=1)
        """
        if self.isEnabledFor(WARNING):
            msg = ' '.join([str(m) for m in msg]) + end
            self._log(WARNING, msg, args, **kwargs)

    def error(self, *msg, args: tuple = (), end='\033[0m\n', **kwargs):
        """
        Log 'msg % args' with severity 'ERROR'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.error("Houston, we have a %s", args=("major problem",), exc_info=1)
        """
        if self.isEnabledFor(ERROR):
            msg = ' '.join([str(m) for m in msg]) + end
            self._log(ERROR, msg, args, **kwargs)

    def exception(self, *msg, args, exc_info=True, **kwargs):
        """
        Convenience method for logging an ERROR with exception information.
        """
        self.error(*msg, args=args, exc_info=exc_info, **kwargs)

    def critical(self, *msg, args: tuple = (), end='\033[0m\n', **kwargs):
        """
        Log 'msg % args' with severity 'CRITICAL'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.critical("Houston, we have a %s", args=("major disaster",), exc_info=1)
        """
        if self.isEnabledFor(CRITICAL):
            msg = ' '.join([str(m) for m in msg]) + end
            self._log(CRITICAL, msg, args, **kwargs)

    fatal = critical

    def print(self, *msg, args: tuple = (), end='\033[0m\n', **kwargs):
        """
        Log 'msg % args'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.print("Houston, we have a %s", args=("major disaster",), exc_info=1)
        """
        msg = ' '.join([str(m) for m in msg]) + end
        self._log(self.level if self.level >= NOTSET else NOTSET, msg, args, **kwargs)

    def print_progress(self, progress: float, total: float):
        """
        Log progress.
        """
        self.progress_bar(progress, total)

    @property
    def progress_bar(self):
        if not self._progress_bar:
            from log21.ProgressBar import ProgressBar
            self._progress_bar = ProgressBar(logger=self)
        return self._progress_bar

    @progress_bar.setter
    def progress_bar(self, value: '_log21.ProgressBar'):
        self._progress_bar = value



# log21.StreamHandler.py
# CodeWriter21

import os as _os
import re as _re
from logging import StreamHandler as _StreamHandler
from log21.Colors import ansi_esc, get_colors

__all__ = ['IS_WINDOWS', 'ColorizingStreamHandler', 'StreamHandler']

IS_WINDOWS = _os.name == 'nt'

if IS_WINDOWS:
    import ctypes


class StreamHandler(_StreamHandler):
    terminator = ''

    def __init__(self, handle_carriage_return: bool = True, handle_new_line: bool = True, stream=None):
        self.HandleCR = handle_carriage_return
        self.HandleNL = handle_new_line
        super().__init__(stream=stream)

    def check_cr(self, record):
        if record.msg:
            if '\r' in record.msg[:-1]:
                file_descriptor = getattr(self.stream, 'fileno', None)
                if file_descriptor:
                    file_descriptor = file_descriptor()
                    if file_descriptor in (1, 2):  # stdout or stderr
                        self.stream.write('\r' + (' ' * (_os.get_terminal_size(file_descriptor)[0] - 1)) + '\r')
                        index = record.msg.rfind('\r')
                        find = _re.compile(r'(\x1b\[(?:\d+(?:;(?:\d+))*)m)')
                        record.msg = get_colors(*find.split(record.msg[:index])) + record.msg[index + 1:]

    def check_nl(self, record):
        if record.msg:
            while record.msg[0] == '\n':
                file_descriptor = getattr(self.stream, 'fileno', None)
                if file_descriptor:
                    file_descriptor = file_descriptor()
                    if file_descriptor in (1, 2):  # stdout or stderr
                        self.stream.write('\n')
                        record.msg = record.msg[1:]

    def emit(self, record):
        if self.HandleCR:
            self.check_cr(record)
        if self.HandleNL:
            self.check_nl(record)
        super().emit(record)


# A stream handler that supports colorizing.
class ColorizingStreamHandler(StreamHandler):
    def emit(self, record):
        try:
            if self.HandleCR:
                self.check_cr(record)
            if self.HandleNL:
                self.check_nl(record)
            msg = self.format(record)
            if IS_WINDOWS:
                self.convert_and_write(msg)
                self.convert_and_write(self.terminator)
            else:
                self.write(msg)
                self.write(self.terminator)
            self.flush()
        except Exception:
            self.handleError(record)

    # Writes colorized text to the Windows console.
    def convert_and_write(self, message):
        nt_color_map = {
            30: 0,  # foreground: black   - 0b00000000
            31: 4,  # foreground: red     - 0b00000100
            32: 2,  # foreground: green   - 0b00000010
            33: 6,  # foreground: yellow  - 0b00000110
            34: 1,  # foreground: blue    - 0b00000001
            35: 5,  # foreground: magenta - 0b00000101
            36: 3,  # foreground: cyan    - 0b00000011
            37: 7,  # foreground: white   - 0b00000111
            40: 0,  # background: black   - 0b00000000 = 0 << 4
            41: 64,  # background: red     - 0b01000000 = 4 << 4
            42: 32,  # background: green   - 0b00100000 = 2 << 4
            43: 96,  # background: yellow  - 0b01100000 = 6 << 4
            44: 16,  # background: blue    - 0b00010000 = 1 << 4
            45: 80,  # background: magenta - 0b01010000 = 5 << 4
            46: 48,  # background: cyan    - 0b00110000 = 3 << 4
            47: 112,  # background: white   - 0b01110000 = 7 << 4
            90: 8,  # foreground: gray          - 0b00001000
            91: 12,  # foreground: light red     - 0b00001100
            92: 10,  # foreground: light green   - 0b00001010
            93: 14,  # foreground: light yellow  - 0b00001110
            94: 9,  # foreground: light blue    - 0b00001001
            95: 13,  # foreground: light magenta - 0b00001101
            96: 11,  # foreground: light cyan    - 0b00001011
            97: 15,  # foreground: light white   - 0b00001111
            100: 128,  # background: gray          - 0b10000000 = 8  << 4
            101: 192,  # background: light red     - 0b11000000 = 12 << 4
            102: 160,  # background: light green   - 0b10100000 = 10 << 4
            103: 224,  # background: light yellow  - 0b11100000 = 14 << 4
            104: 144,  # background: light blue    - 0b10010000 = 9  << 4
            105: 208,  # background: light magenta - 0b11010000 = 13 << 4
            106: 176,  # background: light cyan    - 0b10110000 = 11 << 4
            107: 240,  # background: light white   - 0b11110000 = 15 << 4
            2: 8,
            0: 7
        }

        parts = ansi_esc.split(message)
        win_handle = None
        file_descriptor = getattr(self.stream, 'fileno', None)

        if file_descriptor:
            file_descriptor = file_descriptor()

            if file_descriptor in (1, 2):  # stdout or stderr
                win_handle = ctypes.windll.kernel32.GetStdHandle(-10 - file_descriptor)

        while parts:
            text = parts.pop(0)

            if text:
                self.write(text)

            if parts:
                params = parts.pop(0)

                if win_handle is not None:
                    params = [int(p) for p in params.split(';')]
                    color = 0

                    for p in params:
                        if p in nt_color_map:
                            color |= nt_color_map[p]
                        else:
                            pass  # error condition ignored

                    ctypes.windll.kernel32.SetConsoleTextAttribute(win_handle, color)

    # Writes the message to the console.
    def write(self, message):
        self.stream.write(message)
        self.flush()



# log21.Argparse.py
# CodeWriter21

import re as _re
import sys as _sys
import log21 as _log21
import argparse as _argparse
from typing import Dict as _Dict
from gettext import gettext as _gettext
from textwrap import TextWrapper as _TextWrapper
from log21.Colors import get_colors as _gc
from log21.Formatters import DecolorizingFormatter as _Formatter

__all__ = ['ColorizingArgumentParser', 'ColorizingHelpFormatter', 'ColorizingTextWrapper']


class ColorizingHelpFormatter(_argparse.HelpFormatter):
    colors = {
        'usage': 'Cyan',
        'brackets': 'LightRed',
        'switches': 'LightCyan',
        'values': 'Green',
        'colons': 'LightRed',
        'commas': 'LightRed',
        'section headers': 'LightGreen',
        'help': 'LightWhite',
        'choices': 'LightGreen'
    }

    def __init__(self, prog, indent_increment=2, max_help_position=24, width=None, colors: _Dict[str, str] = None):
        super().__init__(prog, indent_increment, max_help_position, width)
        if colors:
            for key, value in colors.items():
                if key in self.colors:
                    self.colors[key] = value

    class _Section(object):
        def __init__(self, formatter, parent, heading=None):
            self.formatter = formatter
            self.parent = parent
            self.heading = heading
            self.items = []

        def format_help(self):
            # format the indented section
            if self.parent is not None:
                self.formatter._indent()
            join = self.formatter._join_parts
            item_help = join([func(*args) for func, args in self.items])
            if self.parent is not None:
                self.formatter._dedent()

            # return nothing if the section was empty
            if not item_help:
                return ''

            # add the heading if the section was non-empty
            if self.heading is not _argparse.SUPPRESS and self.heading is not None:
                current_indent = self.formatter._current_indent
                heading = '%*s%s' % (current_indent, '', self.heading) + _gc(self.formatter.colors['colons']) + \
                          ':\033[0m\n'
            else:
                heading = ''

            # join the section-initial newline, the heading and the help
            return join(['\n', heading, _gc(self.formatter.colors['help']), item_help, '\n'])

    def _add_item(self, func, args):
        self._current_section.items.append((func, args))

    def _fill_text(self, text, width, indent):
        text = self._whitespace_matcher.sub(' ', text).strip()
        return ColorizingTextWrapper(width=width, initial_indent=indent, subsequent_indent=indent).fill(text)

    def _split_lines(self, text, width):
        text = self._whitespace_matcher.sub(' ', text).strip()
        return ColorizingTextWrapper(width=width).wrap(text)

    def start_section(self, heading):
        self._indent()
        section = self._Section(self, self._current_section,
                                _gc(self.colors['section headers']) + str(heading) + '\033[0m')
        self._add_item(section.format_help, [])
        self._current_section = section

    def _format_action(self, action):
        # determine the required width and the entry label
        help_position = min(self._action_max_length + 2,
                            self._max_help_position)
        help_width = max(self._width - help_position, 11)
        action_width = help_position - self._current_indent - 2
        action_header = _gc('rst') + self._format_action_invocation(action)

        indent_first = 0
        # no help; start on same line and add a final newline
        if not action.help:
            action_header = self._current_indent * ' ' + action_header + '\n'
        # short action name; start on the same line and pad two spaces
        elif len(action_header) <= action_width:
            action_header = '%*s%-*s  ' % (self._current_indent, '', action_width, action_header)
        # long action name; start on the next line
        else:
            action_header = self._current_indent * ' ' + action_header + '\n'
            indent_first = help_position

        # collect the pieces of the action help
        parts = [action_header]

        # if there was help for the action, add lines of help text
        if action.help:
            help_text = _gc(self.colors['help']) + self._expand_help(action)
            help_lines = self._split_lines(help_text, help_width)
            parts.append('%*s%s\n' % (indent_first, '', help_lines[0]))
            for line in help_lines[1:]:
                parts.append('%*s%s\n' % (help_position, '', line))

        # or add a newline if the description doesn't end with one
        elif not action_header.endswith('\n'):
            parts.append('\n')

        # if there are any sub-actions, add their help as well
        for subaction in self._iter_indented_subactions(action):
            parts.append(self._format_action(subaction))

        # return a single string
        return self._join_parts(parts)

    # modified upstream code, not going to refactor for complexity.
    def _format_usage(self, usage, actions, groups, prefix):  # noqa: C901
        if prefix is None:
            prefix = _gettext('usage: ')

        # if usage is specified, use that
        if usage is not None:
            usage = usage % dict(prog=self._prog)

        # if no optionals or positionals are available, usage is just prog
        elif usage is None and not actions:
            usage = '%(prog)s' % dict(prog=self._prog)

        # if optionals and positionals are available, calculate usage
        elif usage is None:
            prog = '%(prog)s' % dict(prog=self._prog)

            # split optionals from positionals
            optionals = []
            positionals = []
            for action in actions:
                if action.option_strings:
                    optionals.append(action)
                else:
                    positionals.append(action)

            # build full usage string
            action_usage = self._format_actions_usage(optionals + positionals, groups)
            usage = ' '.join([s for s in [prog, action_usage] if s])

            # wrap the usage parts if it's too long
            text_width = self._width - self._current_indent
            if len(prefix) + len(_Formatter.decolorize(usage)) > text_width:

                # break usage into wrappable parts
                part_regexp = r'\(.*?\)+|\[.*?\]+|\S+'
                opt_usage = self._format_actions_usage(optionals, groups)
                pos_usage = self._format_actions_usage(positionals, groups)
                opt_parts = _re.findall(part_regexp, opt_usage)
                pos_parts = _re.findall(part_regexp, pos_usage)
                assert ' '.join(opt_parts) == opt_usage
                assert ' '.join(pos_parts) == pos_usage

                # helper for wrapping lines
                def get_lines(parts, indent, prefix=None):
                    lines = []
                    line = []
                    if prefix is not None:
                        line_len = len(prefix) - 1
                    else:
                        line_len = len(indent) - 1
                    for part in parts:
                        if line_len + 1 + len(_Formatter.decolorize(part)) > text_width and line:
                            lines.append(indent + ' '.join(line))
                            line = []
                            line_len = len(indent) - 1
                        line.append(part)
                        line_len += len(_Formatter.decolorize(part)) + 1
                    if line:
                        lines.append(indent + ' '.join(line))
                    if prefix is not None:
                        lines[0] = lines[0][len(indent):]
                    return lines

                # if prog is short, follow it with optionals or positionals
                len_prog = len(_Formatter.decolorize(prog))
                if len(prefix) + len_prog <= 0.75 * text_width:
                    indent = ' ' * (len(prefix) + len_prog + 1)
                    if opt_parts:
                        lines = get_lines([prog] + opt_parts, indent, prefix)
                        lines.extend(get_lines(pos_parts, indent))
                    elif pos_parts:
                        lines = get_lines([prog] + pos_parts, indent, prefix)
                    else:
                        lines = [prog]

                # if prog is long, put it on its own line
                else:
                    indent = ' ' * len(prefix)
                    parts = opt_parts + pos_parts
                    lines = get_lines(parts, indent)
                    if len(lines) > 1:
                        lines = []
                        lines.extend(get_lines(opt_parts, indent))
                        lines.extend(get_lines(pos_parts, indent))
                    lines = [prog] + lines

                # join lines into usage
                usage = '\n'.join(lines)

        # prefix with 'usage:'
        return prefix + usage + '\n\n'

    def _format_actions_usage(self, actions: list, groups):
        # find group indices and identify actions in groups
        group_actions = set()
        inserts = {}
        for group in groups:
            try:
                start = actions.index(group._group_actions[0])
            except ValueError:
                continue
            else:
                end = start + len(group._group_actions)
                if actions[start:end] == group._group_actions:
                    for action in group._group_actions:
                        group_actions.add(action)
                    if not group.required:
                        if start in inserts:
                            inserts[start] += ' ['
                        else:
                            inserts[start] = '['
                        if end in inserts:
                            inserts[end] += ']'
                        else:
                            inserts[end] = ']'
                    else:
                        if start in inserts:
                            inserts[start] += ' ('
                        else:
                            inserts[start] = '('
                        if end in inserts:
                            inserts[end] += ')'
                        else:
                            inserts[end] = ')'
                    for i in range(start + 1, end):
                        inserts[i] = '|'

        # collect all actions format strings
        parts = []
        for i, action in enumerate(actions):

            # suppressed arguments are marked with None
            # remove | separators for suppressed arguments
            if action.help is _argparse.SUPPRESS:
                parts.append(None)
                if inserts.get(i) == '|':
                    inserts.pop(i)
                elif inserts.get(i + 1) == '|':
                    inserts.pop(i + 1)

            # produce all arg strings
            elif not action.option_strings:
                default = self._get_default_metavar_for_positional(action)
                part = self._format_args(action, default)

                # if it's in a group, strip the outer []
                if action in group_actions:
                    if part[0] == '[' and part[-1] == ']':
                        part = part[1:-1]

                # add the action string to the list
                parts.append(part)

            # produce the first way to invoke the option in brackets
            else:
                option_string = action.option_strings[0]

                # if the Optional doesn't take a value, format is:
                #    -s or --long
                if action.nargs == 0:
                    part = _gc(self.colors['switches']) + action.format_usage()

                # if the Optional takes a value, format is:
                #    -s ARGS or --long ARGS
                else:
                    default = self._get_default_metavar_for_optional(action)
                    args_string = self._format_args(action, default)
                    part = _gc(self.colors['switches']) + '%s %s%s' % (option_string, _gc(self.colors['values']),
                                                                       args_string)

                # make it look optional if it's not required or in a group
                if not action.required and action not in group_actions:
                    part = _gc(self.colors['brackets']) + '[' + part + _gc(
                        self.colors['brackets']) + ']\033[0m'

                # add the action string to the list
                parts.append(part)

        # insert things at the necessary indices
        for i in sorted(inserts, reverse=True):
            parts[i:i] = [inserts[i]]

        # join all the action items with spaces
        text = ' '.join([item for item in parts if item is not None])

        # clean up separators for mutually exclusive groups
        open = r'[\[(]'
        close = r'[\])]'
        text = _re.sub(r'(%s) ' % open, r'\1', text)
        text = _re.sub(r' (%s)' % close, r'\1', text)
        text = _re.sub(r'%s *%s' % (open, close), r'', text)
        text = _re.sub(r'\(([^|]*)\)', r'\1', text)
        text = text.strip()

        # return the text
        return text

    def _format_action_invocation(self, action):
        if not action.option_strings:
            default = self._get_default_metavar_for_positional(action)
            metavar, = self._metavar_formatter(action, default)(1)
            return metavar

        else:
            parts = []

            # if the Optional doesn't take a value, format is:
            #    -s, --long
            if action.nargs == 0:
                for option_string in action.option_strings:
                    parts.append(_gc(self.colors['switches']) + option_string)

            # if the Optional takes a value, format is:
            #    -s ARGS, --long ARGS
            else:
                default = self._get_default_metavar_for_optional(action)
                args_string = self._format_args(action, default)
                for option_string in action.option_strings:
                    parts.append(_gc(self.colors['switches']) + '%s %s%s' % (option_string, _gc(self.colors['values']),
                                                                             args_string))

            return _gc(self.colors['commas']) + ', '.join(parts)

    def _metavar_formatter(self, action, default_metavar):
        if action.metavar is not None:
            result = action.metavar
        elif action.choices is not None:
            choice_strs = [str(choice) for choice in action.choices]
            result = _gc(self.colors['brackets']) + '{ ' + (_gc(self.colors['commas']) + ', ').join(
                _gc(self.colors['choices']) + choice_str for choice_str in choice_strs) + \
                     _gc(self.colors['brackets']) + ' }'
        else:
            result = default_metavar

        def format(tuple_size):
            if isinstance(result, tuple):
                return result
            else:
                return (result,) * tuple_size

        return format


class ColorizingTextWrapper(_TextWrapper):
    # modified upstream code, not going to refactor for complexity.
    def _wrap_chunks(self, chunks):  # noqa: C901
        """_wrap_chunks(chunks : [string]) -> [string]

        Wrap a sequence of text chunks and return a list of lines of
        length 'self.width' or less.  (If 'break_long_words' is false,
        some lines may be longer than this.)  Chunks correspond roughly
        to words and the whitespace between them: each chunk is
        indivisible (modulo 'break_long_words'), but a line break can
        come between any two chunks.  Chunks should not have internal
        whitespace; i.e. a chunk is either all whitespace or a "word".
        Whitespace chunks will be removed from the beginning and end of
        lines, but apart from that whitespace is preserved.
        """
        lines = []
        if self.width <= 0:
            raise ValueError("invalid width %r (must be > 0)" % self.width)
        if self.max_lines is not None:
            if self.max_lines > 1:
                indent = self.subsequent_indent
            else:
                indent = self.initial_indent
            if len(indent) + len(self.placeholder.lstrip()) > self.width:
                raise ValueError("placeholder too large for max width")

        # Arrange in reverse order so items can be efficiently popped
        # from a stack of chucks.
        chunks.reverse()

        while chunks:

            # Start the list of chunks that will make up the current line.
            # cur_len is just the length of all the chunks in cur_line.
            cur_line = []
            cur_len = 0

            # Figure out which static string will prefix this line.
            if lines:
                indent = self.subsequent_indent
            else:
                indent = self.initial_indent

            # Maximum width for this line.
            width = self.width - len(indent)

            # First chunk on the line is whitespace -- drop it, unless this
            # is the very beginning of the text (i.e. no lines started yet).
            if self.drop_whitespace and _Formatter.decolorize(chunks[-1]).strip() == '' and lines:
                del chunks[-1]

            while chunks:
                # modified upstream code, not going to refactor for ambiguous variable name.
                l = len(_Formatter.decolorize(chunks[-1]))  # noqa: E741

                # Can at least squeeze this chunk onto the current line.
                # Modified upstream code, not going to refactor for ambiguous variable name.
                if cur_len + l <= width:  # noqa: E741
                    cur_line.append(chunks.pop())
                    cur_len += l
                # Nope, this line is full.
                else:
                    break

            # The current line is full, and the next chunk is too big to
            # fit on *any* line (not just this one).
            if chunks and len(_Formatter.decolorize(chunks[-1])) > width:
                self._handle_long_word(chunks, cur_line, cur_len, width)
                cur_len = sum(map(len, cur_line))

            # If the last chunk on this line is all whitespace, drop it.
            if self.drop_whitespace and cur_line and _Formatter.decolorize(cur_line[-1]).strip() == '':
                cur_len -= len(_Formatter.decolorize(cur_line[-1]))
                del cur_line[-1]

            if cur_line:
                if (self.max_lines is None or
                        len(lines) + 1 < self.max_lines or
                        (not chunks or
                         self.drop_whitespace and
                         len(chunks) == 1 and
                         not chunks[0].strip()) and cur_len <= width):
                    # Convert current line back to a string and store it in
                    # list of all lines (return value).
                    lines.append(indent + ''.join(cur_line))
                else:
                    while cur_line:
                        if _Formatter.decolorize(cur_line[-1]).strip() and cur_len + len(
                                self.placeholder) <= width:
                            cur_line.append(self.placeholder)
                            lines.append(indent + ''.join(cur_line))
                            break
                        cur_len -= len(_Formatter.decolorize(cur_line[-1]))
                        del cur_line[-1]
                    else:
                        if lines:
                            prev_line = lines[-1].rstrip()
                            if len(_Formatter.decolorize(prev_line)) + len(
                                    self.placeholder) <= self.width:
                                lines[-1] = prev_line + self.placeholder
                                break
                        lines.append(indent + self.placeholder.lstrip())
                    break

        return lines


class ColorizingArgumentParser(_argparse.ArgumentParser):
    def __init__(self, formatter_class=ColorizingHelpFormatter, colors: _Dict[str, str] = None, **kwargs):
        self.logger = _log21.Logger('ArgumentParser')
        self.colors = colors
        super().__init__(formatter_class=formatter_class, **kwargs)

    def _print_message(self, message, file=None):
        if message:
            self.logger.handlers.clear()
            handler = _log21.ColorizingStreamHandler(stream=file)
            self.logger.addHandler(handler)
            self.logger.info(message + _gc('rst'))

    def exit(self, status=0, message=None):
        if message:
            self._print_message(_gc('lr') + message + _gc('rst'), _sys.stderr)
        _sys.exit(status)

    def error(self, message):
        self.print_usage(_sys.stderr)
        args = {'prog': self.prog, 'message': message}
        self.exit(2, _gettext(f'%(prog)s: {_gc("r")}error{_gc("lr")}:{_gc("rst")} %(message)s\n') % args)

    def _get_formatter(self):
        if hasattr(self.formatter_class, 'colors'):
            return self.formatter_class(prog=self.prog, colors=self.colors)
        else:
            return self.formatter_class(prog=self.prog)



# log21.PPrint.py
# CodeWriter21

import re as _re
import sys as _sys
import types as _types
import collections as _collections
import dataclasses as _dataclasses

from typing import Dict as _Dict, Union as _Union
from pprint import PrettyPrinter as _PrettyPrinter

from log21.Colors import get_colors as _gc

_builtin_scalars = frozenset({str, bytes, bytearray, float, complex, bool, type(None)})


def _recursion(obj):
    return f"<Recursion on {type(obj).__name__} with id={id(obj)}>"


def _safe_tuple(t):
    """Helper function for comparing 2-tuples"""
    return _SafeKey(t[0]), _SafeKey(t[1])


def _wrap_bytes_repr(obj, width, allowance):
    current = b''
    last = len(obj) // 4 * 4
    for i in range(0, len(obj), 4):
        part = obj[i: i + 4]
        candidate = current + part
        if i == last:
            width -= allowance
        if len(repr(candidate)) > width:
            if current:
                yield repr(current)
            current = part
        else:
            current = candidate
    if current:
        yield repr(current)


class _SafeKey:
    """Helper function for key functions when sorting unorderable objects.

    The wrapped-object will fallback to a Py2.x style comparison for
    unorderable types (sorting first comparing the type name and then by
    the obj ids).  Does not work recursively, so dict.items() must have
    _safe_key applied to both the key and the value.

    """

    __slots__ = ['obj']

    def __init__(self, obj):
        self.obj = obj

    def __lt__(self, other):
        try:
            return self.obj < other.obj
        except TypeError:
            return (str(type(self.obj)), id(self.obj)) < (str(type(other.obj)), id(other.obj))


class PrettyPrinter(_PrettyPrinter):
    signs_colors: _Dict[str, str] = {
        'square-brackets': _gc('LightCyan'),
        'curly-braces': _gc('LightBlue'),
        'parenthesis': _gc('LightGreen'),
        'comma': _gc('LightRed'),
        'colon': _gc('LightRed'),
        '...': _gc('LightMagenta'),
        'data': _gc('Green')
    }

    def __init__(self, indent=1, width=80, depth=None, stream=None, signs_colors: _Dict[str, str] = None, *,
                 compact=False, sort_dicts=True, underscore_numbers=False):
        super().__init__(indent=indent, width=width, depth=depth, stream=stream, compact=compact,
                         sort_dicts=sort_dicts, underscore_numbers=underscore_numbers)
        self._depth = depth
        self._indent_per_level = indent
        self._width = width
        if stream is not None:
            self._stream = stream
        else:
            self._stream = _sys.stdout
        self._compact = bool(compact)
        self._sort_dicts = sort_dicts
        self._underscore_numbers = underscore_numbers
        if signs_colors:
            for sign, color in signs_colors.items():
                self.signs_colors[sign.lower()] = _gc(color)

    def _format(self, obj, stream, indent, allowance, context, level):
        objid = id(obj)
        if objid in context:
            stream.write(_recursion(obj))
            self._recursive = True
            self._readable = False
            return
        rep = self._repr(obj, context, level)
        max_width = self._width - indent - allowance
        if len(rep) > max_width:
            p = self._dispatch.get(type(obj).__repr__, None)
            if p is not None:
                context[objid] = 1
                p(self, obj, stream, indent, allowance, context, level + 1)
                del context[objid]
                return
            elif (_dataclasses.is_dataclass(obj) and
                  not isinstance(obj, type) and
                  obj.__dataclass_params__.repr and
                  # Check dataclass has generated repr method.
                  hasattr(obj.__repr__, "__wrapped__") and
                  "__create_fn__" in obj.__repr__.__wrapped__.__qualname__):
                context[objid] = 1
                self._pprint_dataclass(obj, stream, indent, allowance, context, level + 1)
                del context[objid]
                return
        stream.write(rep)

    def _pprint_dataclass(self, obj, stream, indent, allowance, context, level):
        cls_name = obj.__class__.__name__
        indent += len(cls_name) + 1
        items = [(f.name, getattr(obj, f.name)) for f in _dataclasses.fields(obj) if f.repr]
        stream.write(cls_name + '(')
        self._format_namespace_items(items, stream, indent, allowance, context, level)
        stream.write(')')

    def _repr(self, obj, context, level):
        repr, readable, recursive = self.format(obj, context.copy(),
                                                self._depth, level)
        if not readable:
            self._readable = False
        if recursive:
            self._recursive = True
        return repr

    def _safe_repr(self, obj, context, max_levels, level):
        # Return triple (repr_string, isreadable, isrecursive).
        typ = type(obj)
        if typ in _builtin_scalars:
            return repr(obj), True, False

        r = getattr(typ, "__repr__", None)

        if issubclass(typ, int) and r is int.__repr__:
            if self._underscore_numbers:
                return f"{obj:_d}", True, False
            else:
                return repr(obj), True, False

        if issubclass(typ, dict) and r is dict.__repr__:
            if not obj:
                return self.signs_colors.get('curly-braces') + "{}" + self.signs_colors.get('data'), True, False
            objid = id(obj)
            if max_levels and level >= max_levels:
                return self.signs_colors.get('curly-braces') + "{" + self.signs_colors.get('...') + "..." + \
                       self.signs_colors.get('curly-braces') + "}" + self.signs_colors.get('data'), \
                       False, objid in context
            if objid in context:
                return _recursion(obj), False, True
            context[objid] = 1
            readable = True
            recursive = False
            components = []
            append = components.append
            level += 1
            if self._sort_dicts:
                items = sorted(obj.items(), key=_safe_tuple)
            else:
                items = obj.items()
            for k, v in items:
                krepr, kreadable, krecur = self.format(k, context, max_levels, level)
                vrepr, vreadable, vrecur = self.format(v, context, max_levels, level)
                append(f"{krepr}{self.signs_colors.get('colon')}:{self.signs_colors.get('data')} {vrepr}")
                readable = readable and kreadable and vreadable
                if krecur or vrecur:
                    recursive = True
            del context[objid]
            return self.signs_colors.get('curly-braces') + "{" + self.signs_colors.get('data') + \
                   (self.signs_colors.get('comma') + ", " + self.signs_colors.get('data')).join(components) + \
                   self.signs_colors.get('data'), readable, recursive

        if (issubclass(typ, list) and r is list.__repr__) or \
                (issubclass(typ, tuple) and r is tuple.__repr__):
            if issubclass(typ, list):
                if not obj:
                    return self.signs_colors.get('square-brackets') + "[]" + self.signs_colors.get('data'), True, False
                format_ = self.signs_colors.get('square-brackets') + "[" + self.signs_colors.get('data') + "%s" + \
                          self.signs_colors.get('square-brackets') + "]" + self.signs_colors.get('data')
            elif len(obj) == 1:
                format_ = self.signs_colors.get('parenthesis') + "(" + self.signs_colors.get('data') + "%s" + \
                          self.signs_colors.get('comma') + "," + self.signs_colors.get('parenthesis') + ")" + \
                          self.signs_colors.get('data')
            else:
                if not obj:
                    return self.signs_colors.get('parenthesis') + "()" + self.signs_colors.get('data'), True, False
                format_ = self.signs_colors.get('parenthesis') + "(" + self.signs_colors.get('data') + "%s" + \
                          self.signs_colors.get('parenthesis') + ")" + self.signs_colors.get('data')
            objid = id(obj)
            if max_levels and level >= max_levels:
                return format_ % self.signs_colors.get('...') + "...", False, objid in context
            if objid in context:
                return _recursion(obj), False, True
            context[objid] = 1
            readable = True
            recursive = False
            components = []
            append = components.append
            level += 1
            for o in obj:
                orepr, oreadable, orecur = self.format(o, context, max_levels, level)
                append(orepr)
                if not oreadable:
                    readable = False
                if orecur:
                    recursive = True
            del context[objid]
            return format_ % (self.signs_colors.get('comma') + ", " + self.signs_colors.get('data')).join(components), \
                   readable, recursive

        rep = repr(obj)
        return rep, (rep and not rep.startswith('<')), False

    _dispatch = {}

    def _pprint_dict(self, obj, stream, indent, allowance, context, level):
        write = stream.write
        write(self.signs_colors.get('curly-braces') + '{' + self.signs_colors.get('data'))
        if self._indent_per_level > 1:
            write((self._indent_per_level - 1) * ' ')
        length = len(obj)
        if length:
            if self._sort_dicts:
                items = sorted(obj.items(), key=_safe_tuple)
            else:
                items = obj.items()
            self._format_dict_items(items, stream, indent, allowance + 1,
                                    context, level)
        write(self.signs_colors.get('curly-braces') + '}' + self.signs_colors.get('data'))

    _dispatch[dict.__repr__] = _pprint_dict

    def _pprint_ordered_dict(self, obj, stream, indent, allowance, context, level):
        if not len(obj):
            stream.write(repr(obj))
            return
        cls = obj.__class__
        stream.write(cls.__name__ + self.signs_colors.get('parenthesis') + '(' + self.signs_colors.get('data'))
        self._format(list(obj.items()), stream,
                     indent + len(cls.__name__) + 1, allowance + 1,
                     context, level)
        stream.write(self.signs_colors.get('parenthesis') + ')' + self.signs_colors.get('data'))

    _dispatch[_collections.OrderedDict.__repr__] = _pprint_ordered_dict

    def _pprint_list(self, obj, stream, indent, allowance, context, level):
        stream.write(self.signs_colors.get('square-brackets') + '[' + self.signs_colors.get('data'))
        self._format_items(obj, stream, indent, allowance + 1,
                           context, level)
        stream.write(self.signs_colors.get('square-brackets') + ']' + self.signs_colors.get('data'))

    _dispatch[list.__repr__] = _pprint_list

    def _pprint_tuple(self, obj, stream, indent, allowance, context, level):
        stream.write(self.signs_colors.get('parenthesis') + '(' + self.signs_colors.get('data'))
        endchar = self.signs_colors.get('comma') + ',' + self.signs_colors.get('parenthesis') + ')' + \
                  self.signs_colors.get('data') if len(obj) == 1 else self.signs_colors.get('parenthesis') + ')' + \
                                                                      self.signs_colors.get('data')
        self._format_items(obj, stream, indent, allowance + len(endchar),
                           context, level)
        stream.write(endchar)

    _dispatch[tuple.__repr__] = _pprint_tuple

    def _pprint_set(self, obj, stream, indent, allowance, context, level):
        if not len(obj):
            stream.write(repr(obj))
            return
        typ = obj.__class__
        if typ is set:
            stream.write(self.signs_colors.get('curly-braces') + '{' + self.signs_colors.get('data'))
            endchar = self.signs_colors.get('curly-braces') + '}' + self.signs_colors.get('data')
        else:
            stream.write(typ.__name__ + self.signs_colors.get('parenthesis') + '(' +
                         self.signs_colors.get('curly-braces') + '{' + self.signs_colors.get('data'))
            endchar = self.signs_colors.get('curly-braces') + '}' + self.signs_colors.get('parenthesis') + ')' \
                      + self.signs_colors.get('data')
            indent += len(typ.__name__) + 1
        obj = sorted(obj, key=_SafeKey)
        self._format_items(obj, stream, indent, allowance + len(endchar),
                           context, level)
        stream.write(endchar)

    _dispatch[set.__repr__] = _pprint_set
    _dispatch[frozenset.__repr__] = _pprint_set

    def _pprint_str(self, obj, stream, indent, allowance, context, level):
        write = stream.write
        if not len(obj):
            write(repr(obj))
            return
        chunks = []
        lines = obj.splitlines(True)
        if level == 1:
            indent += 1
            allowance += 1
        max_width1 = max_width = self._width - indent
        rep = ''
        for i, line in enumerate(lines):
            rep = repr(line)
            if i == len(lines) - 1:
                max_width1 -= allowance
            if len(rep) <= max_width1:
                chunks.append(rep)
            else:
                # A list of alternating (non-space, space) strings
                parts = _re.findall(r'\S*\s*', line)
                assert parts
                assert not parts[-1]
                parts.pop()  # drop empty last part
                max_width2 = max_width
                current = ''
                for j, part in enumerate(parts):
                    candidate = current + part
                    if j == len(parts) - 1 and i == len(lines) - 1:
                        max_width2 -= allowance
                    if len(repr(candidate)) > max_width2:
                        if current:
                            chunks.append(repr(current))
                        current = part
                    else:
                        current = candidate
                if current:
                    chunks.append(repr(current))
        if len(chunks) == 1:
            write(rep)
            return
        if level == 1:
            write(self.signs_colors.get('parenthesis') + '(' + self.signs_colors.get('data'))
        for i, rep in enumerate(chunks):
            if i > 0:
                write('\n' + ' ' * indent)
            write(rep)
        if level == 1:
            write(self.signs_colors.get('parenthesis') + ')' + self.signs_colors.get('data'))

    _dispatch[str.__repr__] = _pprint_str

    def _pprint_bytes(self, obj, stream, indent, allowance, context, level):
        write = stream.write
        if len(obj) <= 4:
            write(repr(obj))
            return
        parens = level == 1
        if parens:
            indent += 1
            allowance += 1
            write(self.signs_colors.get('parenthesis') + '(' + self.signs_colors.get('data'))
        delim = ''
        for rep in _wrap_bytes_repr(obj, self._width - indent, allowance):
            write(delim)
            write(rep)
            if not delim:
                delim = '\n' + ' ' * indent
        if parens:
            write(self.signs_colors.get('parenthesis') + ')' + self.signs_colors.get('data'))

    _dispatch[bytes.__repr__] = _pprint_bytes

    def _pprint_bytearray(self, obj, stream, indent, allowance, context, level):
        write = stream.write
        write('bytearray' + self.signs_colors.get('parenthesis') + '(' + self.signs_colors.get('data'))
        self._pprint_bytes(bytes(obj), stream, indent + 10,
                           allowance + 1, context, level + 1)
        write(self.signs_colors.get('parenthesis') + ')' + self.signs_colors.get('data'))

    _dispatch[bytearray.__repr__] = _pprint_bytearray

    def _pprint_mappingproxy(self, obj, stream, indent, allowance, context, level):
        stream.write('mappingproxy' + self.signs_colors.get('parenthesis') + '(' + self.signs_colors.get('data'))
        self._format(obj.copy(), stream, indent + 13, allowance + 1,
                     context, level)
        stream.write(self.signs_colors.get('parenthesis') + ')' + self.signs_colors.get('data'))

    _dispatch[_types.MappingProxyType.__repr__] = _pprint_mappingproxy

    def _pprint_simplenamespace(self, obj, stream, indent, allowance, context, level):
        if type(obj) is _types.SimpleNamespace:
            # The SimpleNamespace repr is "namespace" instead of the class
            # name, so we do the same here. For subclasses; use the class name.
            cls_name = 'namespace'
        else:
            cls_name = obj.__class__.__name__
        indent += len(cls_name) + 1
        items = obj.__dict__.items()
        stream.write(cls_name + self.signs_colors.get('parenthesis') + '(' + self.signs_colors.get('data'))
        self._format_namespace_items(items, stream, indent, allowance, context, level)
        stream.write(self.signs_colors.get('parenthesis') + ')' + self.signs_colors.get('data'))

    _dispatch[_types.SimpleNamespace.__repr__] = _pprint_simplenamespace

    def _format_dict_items(self, items, stream, indent, allowance, context, level):
        write = stream.write
        indent += self._indent_per_level
        delimnl = self.signs_colors.get('comma') + ',\n' + self.signs_colors.get('data') + ' ' * indent
        last_index = len(items) - 1
        for i, (key, ent) in enumerate(items):
            last = i == last_index
            rep = self._repr(key, context, level)
            write(rep)
            write(self.signs_colors.get('colon') + ': ' + self.signs_colors.get('data'))
            self._format(ent, stream, indent + len(rep) + 2,
                         allowance if last else 1,
                         context, level)
            if not last:
                write(delimnl)

    def _format_namespace_items(self, items, stream, indent, allowance, context, level):
        write = stream.write
        delimnl = self.signs_colors.get('comma') + ',\n' + self.signs_colors.get('data') + ' ' * indent
        last_index = len(items) - 1
        for i, (key, ent) in enumerate(items):
            last = i == last_index
            write(key)
            write('=')
            if id(ent) in context:
                # Special-case representation of recursion to match standard
                # recursive dataclass repr.
                write(self.signs_colors.get('...') + "..." + self.signs_colors.get('data'))
            else:
                self._format(ent, stream, indent + len(key) + 1, allowance if last else 1, context, level)
            if not last:
                write(delimnl)

    def _format_items(self, items, stream, indent, allowance, context, level):
        write = stream.write
        indent += self._indent_per_level
        if self._indent_per_level > 1:
            write((self._indent_per_level - 1) * ' ')
        delimnl = self.signs_colors.get('comma') + ',\n' + self.signs_colors.get('data') + ' ' * indent
        delim = ''
        width = max_width = self._width - indent + 1
        it = iter(items)
        try:
            next_ent = next(it)
        except StopIteration:
            return
        last = False
        while not last:
            ent = next_ent
            try:
                next_ent = next(it)
            except StopIteration:
                last = True
                max_width -= allowance
                width -= allowance
            if self._compact:
                rep = self._repr(ent, context, level)
                w = len(rep) + 2
                if width < w:
                    width = max_width
                    if delim:
                        delim = delimnl
                if width >= w:
                    width -= w
                    write(delim)
                    delim = self.signs_colors.get('comma') + ', ' + self.signs_colors.get('data')
                    write(rep)
                    continue
            write(delim)
            delim = delimnl
            self._format(ent, stream, indent, allowance if last else 1, context, level)

    def _pprint_default_dict(self, obj, stream, indent, allowance, context, level):
        if not len(obj):
            stream.write(repr(obj))
            return
        rdf = self._repr(obj.default_factory, context, level)
        cls = obj.__class__
        indent += len(cls.__name__) + 1
        stream.write(cls.__name__ + self.signs_colors.get('parenthesis') + '(' + self.signs_colors.get('data') + rdf +
                     self.signs_colors.get('comma') + ',\n' + self.signs_colors.get('data') + (' ' * indent))
        self._pprint_dict(obj, stream, indent, allowance + 1, context, level)
        stream.write(self.signs_colors.get('parenthesis') + ')' + self.signs_colors.get('data'))

    _dispatch[_collections.defaultdict.__repr__] = _pprint_default_dict

    def _pprint_counter(self, obj, stream, indent, allowance, context, level):
        if not len(obj):
            stream.write(repr(obj))
            return
        cls = obj.__class__
        stream.write(cls.__name__ + self.signs_colors.get('parenthesis') + '(' +
                     self.signs_colors.get('curly-braces') + '{' + self.signs_colors.get('data'))
        if self._indent_per_level > 1:
            stream.write((self._indent_per_level - 1) * ' ')
        items = obj.most_common()
        self._format_dict_items(items, stream,
                                indent + len(cls.__name__) + 1, allowance + 2,
                                context, level)
        stream.write(self.signs_colors.get('curly-braces') + '}' +
                     self.signs_colors.get('parenthesis') + ')' + self.signs_colors.get('data'))

    _dispatch[_collections.Counter.__repr__] = _pprint_counter

    def _pprint_chain_map(self, obj, stream, indent, allowance, context, level):
        if not len(obj.maps):
            stream.write(repr(obj))
            return
        cls = obj.__class__
        stream.write(cls.__name__ + self.signs_colors.get('parenthesis') + '(' + self.signs_colors.get('data'))
        indent += len(cls.__name__) + 1
        for i, m in enumerate(obj.maps):
            if i == len(obj.maps) - 1:
                self._format(m, stream, indent, allowance + 1, context, level)
                stream.write(self.signs_colors.get('parenthesis') + ')' + self.signs_colors.get('data'))
            else:
                self._format(m, stream, indent, 1, context, level)
                stream.write(self.signs_colors.get('comma') + ',\n' + self.signs_colors.get('data') + ' ' * indent)

    _dispatch[_collections.ChainMap.__repr__] = _pprint_chain_map

    def _pprint_deque(self, obj, stream, indent, allowance, context, level):
        if not len(obj):
            stream.write(repr(obj))
            return
        cls = obj.__class__
        stream.write(cls.__name__ + self.signs_colors.get('parenthesis') + '(' + self.signs_colors.get('data'))
        indent += len(cls.__name__) + 1
        stream.write(self.signs_colors.get('square-brackets') + '[' + self.signs_colors.get('data'))
        if obj.maxlen is None:
            self._format_items(obj, stream, indent, allowance + 2,
                               context, level)
            stream.write(self.signs_colors.get('square-brackets') + ']' + self.signs_colors.get('parenthesis') + ')' +
                         self.signs_colors.get('data'))
        else:
            self._format_items(obj, stream, indent, 2,
                               context, level)
            rml = self._repr(obj.maxlen, context, level)
            stream.write(self.signs_colors.get('square-brackets') + ']' + self.signs_colors.get('comma') +
                         ',' + self.signs_colors.get('data') + '\n' + (' ' * indent) + 'maxlen=' + rml +
                         self.signs_colors.get('parenthesis') + ')')

    _dispatch[_collections.deque.__repr__] = _pprint_deque

    def _pprint_user_dict(self, obj, stream, indent, allowance, context, level):
        self._format(obj.data, stream, indent, allowance, context, level - 1)

    _dispatch[_collections.UserDict.__repr__] = _pprint_user_dict

    def _pprint_user_list(self, obj, stream, indent, allowance, context, level):
        self._format(obj.data, stream, indent, allowance, context, level - 1)

    _dispatch[_collections.UserList.__repr__] = _pprint_user_list

    def _pprint_user_string(self, obj, stream, indent, allowance, context, level):
        self._format(obj.data, stream, indent, allowance, context, level - 1)

    _dispatch[_collections.UserString.__repr__] = _pprint_user_string


def pformat(obj, indent=1, width=80, depth=None, signs_colors: _Dict[str, str] = None, *, compact=False,
            sort_dicts=True, underscore_numbers=False):
    """Format a Python object into a pretty-printed representation."""
    return PrettyPrinter(indent=indent, width=width, depth=depth, compact=compact, signs_colors=signs_colors,
                         sort_dicts=sort_dicts, underscore_numbers=underscore_numbers).pformat(obj)




#!/usr/bin/env python3
# setup.py

from setuptools import setup

with open('README.md', 'r') as file:
    long_description = file.read()

DESCRIPTION = 'A simple logging package that helps you log colorized messages in Windows console.'
VERSION = '1.5.10'

setup(
    name='log21',
    version=VERSION,
    url='https://github.com/MPCodeWriter21/log21',
    author='CodeWriter21(Mehrad Pooryoussof)',
    author_email='<CodeWriter21@gmail.com>',
    license='Apache-2.0 License',
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=['webcolors'],
    packages=['log21'],
    keywords=['python', 'log', 'colorize', 'color', 'logging'],
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.10",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows"
    ]
)



# log21.Colors.py
# CodeWriter21

import re as _re
import webcolors as _webcolors
from typing import Union as _Union, Sequence as _Sequence

__all__ = ['Colors', 'get_color', 'get_colors', 'ansi_esc', 'get_color_name', 'closest_color']

# Regex pattern to find ansi colors in message
ansi_esc = _re.compile(r'\x1b\[((?:\d+)(?:;(?:\d+))*)m')


class Colors:
    color_map = {
        'Reset': '\033[0m',
        'Black': '\033[30m',
        'Red': '\033[31m',
        'Green': '\033[32m',
        'Yellow': '\033[33m',
        'Blue': '\033[34m',
        'Magenta': '\033[35m',
        'Cyan': '\033[36m',
        'White': '\033[37m',
        'BackBlack': '\033[40m',
        'BackRed': '\033[41m',
        'BackGreen': '\033[42m',
        'BackYellow': '\033[43m',
        'BackBlue': '\033[44m',
        'BackMagenta': '\033[45m',
        'BackCyan': '\033[46m',
        'BackWhite': '\033[47m',
        'Grey': '\033[90m',
        'LightRed': '\033[91m',
        'LightGreen': '\033[92m',
        'LightYellow': '\033[93m',
        'LightBlue': '\033[94m',
        'LightMagenta': '\033[95m',
        'LightCyan': '\033[96m',
        'LightWhite': '\033[97m',
        'BackGrey': '\033[100m',
        'BackLightRed': '\033[101m',
        'BackLightGreen': '\033[102m',
        'BackLightYellow': '\033[103m',
        'BackLightBlue': '\033[104m',
        'BackLightMagenta': '\033[105m',
        'BackLightCyan': '\033[106m',
        'BackLightWhite': '\033[107m'
    }
    color_map_ = {
        'reset': '\x1b[0m',
        'black': '\x1b[30m',
        'red': '\x1b[31m',
        'green': '\x1b[32m',
        'yellow': '\x1b[33m',
        'blue': '\x1b[34m',
        'magenta': '\x1b[35m',
        'cyan': '\x1b[36m',
        'white': '\x1b[37m',
        'backblack': '\x1b[40m',
        'backred': '\x1b[41m',
        'backgreen': '\x1b[42m',
        'backyellow': '\x1b[43m',
        'backblue': '\x1b[44m',
        'backmagenta': '\x1b[45m',
        'backcyan': '\x1b[46m',
        'backwhite': '\x1b[47m',
        'grey': '\x1b[90m',
        'lightred': '\x1b[91m',
        'lightgreen': '\x1b[92m',
        'lightyellow': '\x1b[93m',
        'lightblue': '\x1b[94m',
        'lightmagenta': '\x1b[95m',
        'lightcyan': '\x1b[96m',
        'lightwhite': '\x1b[97m',
        'backgrey': '\x1b[100m',
        'backlightred': '\x1b[101m',
        'backlightgreen': '\x1b[102m',
        'backlightyellow': '\x1b[103m',
        'backlightblue': '\x1b[104m',
        'backlightmagenta': '\x1b[105m',
        'backlightcyan': '\x1b[106m',
        'backlightwhite': '\x1b[107m',
        'rst': '\x1b[0m',
        'bk': '\x1b[30m',
        'r': '\x1b[31m',
        'g': '\x1b[32m',
        'y': '\x1b[33m',
        'b': '\x1b[34m',
        'm': '\x1b[35m',
        'c': '\x1b[36m',
        'w': '\x1b[37m',
        'bbk': '\x1b[40m',
        'br': '\x1b[41m',
        'bg': '\x1b[42m',
        'by': '\x1b[43m',
        'bb': '\x1b[44m',
        'bm': '\x1b[45m',
        'bc': '\x1b[46m',
        'bw': '\x1b[47m',
        'gr': '\x1b[90m',
        'lr': '\x1b[91m',
        'lg': '\x1b[92m',
        'ly': '\x1b[93m',
        'lb': '\x1b[94m',
        'lm': '\x1b[95m',
        'lc': '\x1b[96m',
        'lw': '\x1b[97m',
        'bgr': '\x1b[100m',
        'blr': '\x1b[101m',
        'blg': '\x1b[102m',
        'bly': '\x1b[103m',
        'blb': '\x1b[104m',
        'blm': '\x1b[105m',
        'blc': '\x1b[106m',
        'blw': '\x1b[107m',
    }
    change_map = {
        'aqua': 'LightCyan',
        'blue': 'LightBlue',
        'fuchsia': 'LightMagenta',
        'lime': 'LightGreen',
        'maroon': 'Red',
        'navy': 'Blue',
        'olive': 'Yellow',
        'purple': 'Magenta',
        'red': 'LightRed',
        'silver': 'Grey',
        'teal': 'Cyan',
        'white': 'LightWhite',
        'yellow': 'LightYellow',
    }


def closest_color(requested_color: _Sequence[int]):
    """
    Takes a color in RGB and returns the name of the closest color to the value.
    Uses the `webcolors.CSS2_HEX_TO_NAMES` dictionary to find the closest color.

    :param requested_color: Sequence[int, int, int]: The input color in RGB.
    :return: str: The name of the closest color.
    """
    min_colors = {}
    for key, name in _webcolors.CSS2_HEX_TO_NAMES.items():
        r_c, g_c, b_c = _webcolors.hex_to_rgb(key)
        rd = (r_c - requested_color[0]) ** 2
        gd = (g_c - requested_color[1]) ** 2
        bd = (b_c - requested_color[2]) ** 2
        min_colors[(rd + gd + bd)] = name
    return min_colors[min(min_colors.keys())]


def get_color_name(color: _Union[str, _Sequence[int], _webcolors.IntegerRGB, _webcolors.HTML5SimpleColor],
                   raise_exceptions: bool = False) -> str:
    """
    Takes a color in RGB format and returns a color name close to the RGB value.
    >>>
    >>> get_color_name('#00FF00')
    'LightGreen'
    >>>
    >>> get_color_name((128, 0, 128))
    'Magenta'
    >>>

    :param color: Union[str, Sequence[int]: The input color. Example: '#00FF00', (128, 0, 128)
    :param raise_exceptions: bool = False: Returns empty string when raise_exceptions is False and an error occurs.
    :raises TypeError
    :return: str: The color name.
    """
    # Makes sure that the input parameters has valid values.
    if not isinstance(color, (str, tuple, _webcolors.IntegerRGB, _webcolors.HTML5SimpleColor)):
        if raise_exceptions:
            raise TypeError('Input color must be a str or Tuple[int, int, int] or webcolors.IntegerRGB or ' +
                            'webcolors.HTML5SimpleColor')
        else:
            return ''
    if isinstance(color, str):
        if color.startswith('#') and len(color) == 7:
            color = _webcolors.hex_to_rgb(color)
        elif color.isdigit() and len(color) == 9:
            color = (int(color[:3]), int(color[3:6]), int(color[6:9]))
        else:
            if raise_exceptions:
                raise TypeError('String color format must be `#0021ff` or `000033255`!')
            else:
                return ''
    if isinstance(color, _Sequence):
        if len(color) == 3:
            if not (isinstance(color[0], int) and isinstance(color[1], int) and isinstance(color[2], int)):
                if raise_exceptions:
                    raise TypeError('Color sequence format must be (int, int, int)!')
                else:
                    return ''
        else:
            if raise_exceptions:
                raise TypeError('Color sequence format must be (int, int, int)!')
            else:
                return ''

    # Looks for the name of the input RGB color.
    try:
        closest_name = _webcolors.rgb_to_name(tuple(color))
    except ValueError:
        closest_name = closest_color(color)
    if closest_name in Colors.change_map:
        closest_name = Colors.change_map[closest_name]
    return closest_name


def get_color(color: _Union[str, _Sequence], raise_exceptions: bool = False) -> str:
    """
    Gets a color name and returns it in ansi format
    >>>
    >>> get_color('LightRed')
    '\x1b[91m'
    >>>
    >>> import log21
    >>> log21.get_logger().info(log21.get_color('Blue') + 'Hello World!')
    [21:21:21] [INFO] Hello World!
    >>> # Note that you must run it yourself to see the colorful result ;D
    >>>

    :param color: color name(Example: Blue)
    :param raise_exceptions: bool = False:
        False: It will return '' instead of raising exceptions when an error occurs.
        True: It may raise TypeError or KeyError
    :raises TypeError: `color` must be str
    :raises KeyError: `color` not found!
    :return: str: an ansi color
    """
    if not isinstance(color, (str, _Sequence)):
        if raise_exceptions:
            raise TypeError('`color` must be str or Sequence!')
        else:
            return ''
    if isinstance(color, _Sequence) and not isinstance(color, str):
        color = get_color_name(color)
        return get_color(color)
    else:
        color = color.lower()
        color = color.replace(' ', '').replace('_', '').replace('-', '')
        color = color.replace('foreground', '').replace('fore', '').replace('ground', '')
        if (color.startswith('#') and len(color) == 7) or (color.isdigit() and len(color) == 9):
            color = get_color_name(color)
            return get_color(color)
        elif color in Colors.color_map_:
            return Colors.color_map_[color]
        elif ansi_esc.match(color):
            return ansi_esc.match(color).group()
        elif color in Colors.change_map:
            return get_color(Colors.change_map[color])
        else:
            if raise_exceptions:
                raise KeyError(f'`{color}` not found!')
            else:
                return ''


def get_colors(*colors: str, raise_exceptions: bool = False) -> str:
    """
    Gets a list of colors and combines them into one.
    >>>
    >>> get_colors('LightCyan')
    '\x1b[96m'
    >>>
    >>> import log21
    >>> log21.get_logger().info(log21.get_colors('Green', 'Background White') + 'Hello World!')
    [21:21:21] [INFO] Hello World!
    >>> # Note that you must run it yourself to see the colorful result ;D
    >>>

    :param colors: Input colors
    :param raise_exceptions: bool = False:
        False: It will return '' instead of raising exceptions when an error occurs.
        True: It may raise TypeError or KeyError
    :raises TypeError: `color` must be str
    :raises KeyError: `color` not found!
    :return: str: a combined color
    """
    output = ''
    for color in colors:
        output += get_color(str(color), raise_exceptions=raise_exceptions)
    parts = ansi_esc.split(output)
    output = '\033['
    for part in parts:
        if part:
            output += part + ';'
    if output.endswith(';'):
        output = output[:-1] + 'm'
        return output
    return ''



# log21.Manager.py
# CodeWriter21

import logging as _logging
from typing import Union as _Union
from log21.Logger import Logger as _loggerClass
from log21.Levels import INFO as _INFO

root = _logging.RootLogger(_INFO)


class Manager(_logging.Manager):
    def __init__(self):
        self.root = root
        self.disable = 0
        self.emittedNoHandlerWarning = False
        self.loggerDict = {}
        self.loggerClass = None
        self.logRecordFactory = None

    def getLogger(self, name: str) -> _Union[_logging.Logger, _loggerClass, None]:
        """
        Takes the name of a logger and if there was a logger with that name in the loggerDict it will return the logger
        otherwise it'll return None.

        :param name: The name of the logger.
        :raises TypeError: A logger name must be a string
        :return:
        """
        if not isinstance(name, str):
            raise TypeError('A logger name must be a string')
        try:
            if name in self.loggerDict:
                rv = self.loggerDict[name]
                if isinstance(rv, _logging.PlaceHolder):
                    rv = (self.loggerClass or _loggerClass)(name)
                    rv.manager = self
                    self.loggerDict[name] = rv
            else:
                return None
        except:
            rv = None
        return rv

    def addLogger(self, name: str, logger) -> None:
        """
        Adds a logger to the loggerDict dictionary.

        :param name: str: The name of the logger.
        :param logger: The logger to save.
        :raises TypeError: A logger name must be a string
        :return: None
        """
        if not isinstance(name, str):
            raise TypeError('A logger name must be a string')
        self.loggerDict[name] = logger



# log21.Formatter.py
# CodeWriter21

import time as _time
from logging import Formatter as __Formatter
from typing import Dict as _Dict, Tuple as _Tuple
from log21.Colors import get_colors as _gc, ansi_esc
from log21.Levels import *

__all__ = ['ColorizingFormatter', 'DecolorizingFormatter']


class _Formatter(__Formatter):
    level_names: _Dict[int, str] = {
        DEBUG: 'DEBUG',
        INFO: 'INFO',
        WARNING: 'WARNING',
        ERROR: 'ERROR',
        CRITICAL: 'CRITICAL'
    }

    def __init__(self, fmt: str = None, datefmt: str = None, style: str = '%', level_names: _Dict[int, str] = None):
        """
        `level_names` usage:
        >>> import log21
        >>> logger = log21.Logger('MyLogger', log21.DEBUG)
        >>> stream_handler = log21.ColorizingStreamHandler()
        >>> formatter = log21.ColorizingFormatter(fmt='[%(levelname)s] %(message)s',
        ...             level_names={log21.DEBUG: ' ', log21.INFO: '+', log21.WARNING: '-', log21.ERROR: '!',
        ...                          log21.CRITICAL: 'X'})
        >>> stream_handler.setFormatter(formatter)
        >>> logger.addHandler(stream_handler)
        >>>
        >>> logger.debug('Just wanna see if this works...')
        [ ] Just wanna see if this works...
        >>> logger.info("FYI: I'm glad somebody read this 8)")
        [+] FYI: I'm glad somebody read this 8)
        >>> logger.warning("Oh no! Something's gonna happen!")
        [-] Oh no! something's gonna happen!
        >>> logger.error('AN ERROR OCCURRED! (told ya ;))')
        [!] AN ERROR OCCURRED! (told ya ;))
        >>> logger.critical('Crashed....')
        [X] Crashed....
        >>>
        >>> # Hope you've enjoyed
        >>>
        """
        super().__init__(fmt=fmt, datefmt=datefmt, style=style)
        if level_names:
            for level, name in level_names.items():
                self.level_names[level] = name

    def format(self, record) -> str:
        record.message = record.getMessage()
        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)

        record.levelname = self.level_names.get(record.levelno, 'NOTSET')

        s = self.formatMessage(record)
        if record.exc_info:
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + record.exc_text
        if record.stack_info:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + self.formatStack(record.stack_info)
        return s


class ColorizingFormatter(_Formatter):
    # Default color values
    level_colors: _Dict[int, _Tuple[str, ...]] = {
        DEBUG: ('lightblue',),
        INFO: ('green',),
        WARNING: ('lightyellow',),
        ERROR: ('light red',),
        CRITICAL: ('background red', 'white')
    }
    time_color: _Tuple[str, ...] = ('lightblue',)
    name_color = pathname_color = filename_color = module_color = func_name_color = thread_name_color = \
        message_color = tuple()

    def __init__(self, fmt: str = None, datefmt: str = None, style: str = '%', level_names: _Dict[int, str] = None,
                 level_colors: _Dict[int, _Tuple[str]] = None,
                 time_color: _Tuple[str, ...] = None, name_color: _Tuple[str, ...] = None,
                 pathname_color: _Tuple[str, ...] = None, filename_color: _Tuple[str, ...] = None,
                 module_color: _Tuple[str, ...] = None, func_name_color: _Tuple[str, ...] = None,
                 thread_name_color: _Tuple[str, ...] = None, message_color: _Tuple[str, ...] = None):
        super().__init__(fmt=fmt, datefmt=datefmt, style=style, level_names=level_names)
        # Checks and sets colors
        if level_colors:
            if type(level_colors) is not dict:
                raise TypeError('`level_colors` must be a dictionary!')
            for level, color in level_colors.items():
                self.level_colors[level] = (_gc(*color),)
        if time_color:
            if type(time_color) is not tuple:
                raise TypeError('`time_color` must be a tuple!')
            self.time_color = time_color
        if name_color:
            if type(name_color) is not tuple:
                raise TypeError('`name_color` must be a tuple!')
            self.name_color = name_color
        if pathname_color:
            if type(pathname_color) is not tuple:
                raise TypeError('`pathname_color` must be a tuple!')
            self.pathname_color = pathname_color
        if filename_color:
            if type(filename_color) is not tuple:
                raise TypeError('`filename_color` must be a tuple!')
            self.filename_color = filename_color
        if module_color:
            if type(module_color) is not tuple:
                raise TypeError('`module_color` must be a tuple!')
            self.module_color = module_color
        if func_name_color:
            if type(func_name_color) is not tuple:
                raise TypeError('`func_name_color` must be a tuple!')
            self.func_name_color = func_name_color
        if thread_name_color:
            if type(thread_name_color) is not tuple:
                raise TypeError('`thread_name_color` must be a tuple!')
            self.thread_name_color = thread_name_color
        if message_color:
            if type(message_color) is not tuple:
                raise TypeError('`message_color` must be a tuple!')
            self.message_color = message_color

    def format(self, record) -> str:
        """
        Colorizes the record and returns the formatted message.

        :param record:
        :return: str
        """
        record.message = record.getMessage()
        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)

        record.levelname = self.level_names.get(record.levelno, 'NOTSET')

        record = self.colorize(record)

        s = self.formatMessage(record)
        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + record.exc_text
        if record.stack_info:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + self.formatStack(record.stack_info)
        return s

    def colorize(self, record):
        """
        Colorizes the record attributes.

        :param record:
        :return: colorized record
        """
        reset = '\033[0m'

        if hasattr(record, 'asctime'):
            record.asctime = _gc(*self.time_color) + record.asctime + reset
        if hasattr(record, 'levelno'):
            record.levelname = _gc(*self.level_colors.get(int(record.levelno), ('lw',))) + \
                               getattr(record, 'levelname', 'NOTSET') + reset
        if hasattr(record, 'name'):
            record.name = _gc(*self.name_color) + str(record.name) + reset
        if hasattr(record, 'pathname'):
            record.pathname = _gc(*self.pathname_color) + record.pathname + reset
        if hasattr(record, 'filename'):
            record.filename = _gc(*self.filename_color) + record.filename + reset
        if hasattr(record, 'module'):
            record.module = _gc(*self.module_color) + record.module + reset
        if hasattr(record, 'funcName'):
            record.funcName = _gc(*self.func_name_color) + record.funcName + reset
        if hasattr(record, 'threadName'):
            record.threadName = _gc(*self.thread_name_color) + record.threadName + reset
        if hasattr(record, 'message'):
            record.message = _gc(*self.message_color) + record.message + reset

        return record


class DecolorizingFormatter(_Formatter):
    def formatTime(self, record, datefmt=None):
        """
        Returns the creation time of the specified LogRecord as formatted text.
        """
        ct = self.converter(int(record.created))
        if datefmt:
            s = _time.strftime(datefmt, ct)
        else:
            t = _time.strftime(self.default_time_format, ct)
            s = self.default_msec_format % (t, record.msecs)
        return s

    def format(self, record) -> str:
        """
        Decolorizes the record and returns the formatted message.

        :param record:
        :return: str
        """
        return self.decolorize(super().format(record))

    @staticmethod
    def decolorize(text: str):
        """
        Removes all ansi colors in the text.

        :param text: str: Input text
        :return: str: decolorized text
        """

        return ansi_esc.sub('', text)



# log21.ProgressBar.py
# CodeWriter21

import os as _os

import log21 as _log21
from log21.Logger import Logger as _Logger
from log21.StreamHandler import ColorizingStreamHandler as _ColorizingStreamHandler
from log21.Colors import get_colors as _gc

_logger = _Logger('ProgressBar')
_logger.addHandler(_ColorizingStreamHandler())

__all__ = ['ProgressBar']


class ProgressBar:
    colors = {
        'progress in-progress': _gc('LightYellow'),
        'progress complete': _gc('LightGreen'),
        'progress failed': _gc('LightRed'),
        'percentage in-progress': _gc('LightBlue'),
        'percentage complete': _gc('LightCyan'),
        'percentage failed': _gc('LightRed'),
        'prefix-color in-progress': _gc('Yellow'),
        'prefix-color complete': _gc('Green'),
        'prefix-color failed': _gc('Red'),
        'suffix-color in-progress': _gc('Yellow'),
        'suffix-color complete': _gc('Green'),
        'suffix-color failed': _gc('Red'),
        'reset-color': _gc('Reset'),
    }
    spinner = ['|', '/', '-', '\\']

    def __init__(self, *args, width: int = None, show_percentage: bool = True, prefix: str = '|', suffix: str = '|',
                 fill: str = '█', empty: str = ' ', colors: dict = None, logger: '_log21.Logger' = _logger):
        """
        Example:
            >>> pb = ProgressBar(width=20, show_percentage=False, prefix='[', suffix=']', fill='=', empty='-')
            >>> pb(0, 10)
            [/-----------------]
            >>> pb(1, 10)
            [==----------------]
            >>> pb(2, 10)
            [====\-------------]
            >>>
            >>> # A better example
            >>> import time
            >>> pb = ProgressBar()
            >>> for i in range(500):
            >>>     pb(i + 1, 500)
            >>>     time.sleep(0.01)
            |████████████████████████████████████████████████████████████████████████████████████████████| 100%
            >>> # Of course, You should try it yourself to see the progress! XD
            >>>

        :param args: Prevents the use of positional arguments
        :param width: The width of the progress bar
        :param show_percentage: Whether to show the percentage of the progress
        :param prefix: The prefix of the progress bar
        :param suffix: The suffix of the progress bar
        :param fill: The fill character of the progress bar
        :param empty: The empty character of the progress bar
        :param colors: The colors of the progress bar
        :param logger: The logger to use
        """
        self.width = width if width else _os.get_terminal_size().columns - 1
        if self.width < 3:
            raise ValueError('`width` must be greater than 1')
        if not isinstance(fill, str):
            raise TypeError('`fill` must be a string')
        if not isinstance(empty, str):
            raise TypeError('`empty` must be a string')
        if not isinstance(prefix, str):
            raise TypeError('`prefix` must be a string')
        if not isinstance(suffix, str):
            raise TypeError('`suffix` must be a string')
        if len(fill) != 1:
            raise ValueError('`fill` must be a single character')
        if len(empty) != 1:
            raise ValueError('`empty` must be a single character')

        self.fill = fill
        self.empty = empty
        self.prefix = prefix
        self.suffix = suffix
        self.show_percentage = show_percentage
        if colors:
            for key, value in colors.items():
                self.colors[key] = value
        self.logger = logger
        self.i = 0

    def get_bar(self, progress: float, total: float):
        if progress == total:
            return self.progress_complete()
        elif progress > total or progress < 0:
            return self.progress_failed(progress, total)
        else:
            return self.progress_in_progress(progress, total)

    def progress_in_progress(self, progress: float, total: float):
        percentage = round(progress / total * 100, 2)
        percentage_str = f' {percentage}%' if self.show_percentage else ''
        fill_length = round(progress / total * (self.width - len(percentage_str) - len(self.prefix) - len(self.suffix)))
        empty_length = (self.width - fill_length - len(percentage_str) - len(self.prefix) - len(self.suffix))

        if self.i >= 3:
            self.i = 0
        else:
            self.i += 1
        spinner_char = self.spinner[self.i] if empty_length > 0 else ''

        bar = '\r' + self.colors['prefix-color in-progress'] + self.prefix + \
              self.colors['progress in-progress'] + self.fill * fill_length + spinner_char + \
              self.empty * (empty_length - 1) + \
              self.colors['suffix-color in-progress'] + self.suffix

        return f'{bar}{self.colors["percentage in-progress"]}{percentage_str}' + self.colors['reset-color']

    def progress_complete(self):
        percentage_str = ' 100%' if self.show_percentage else ''
        bar_length = self.width - len(percentage_str) - len(self.prefix) - len(self.suffix)
        bar = '\r' + self.colors['prefix-color complete'] + self.prefix + \
              self.colors['progress complete'] + self.fill * bar_length + \
              self.colors['suffix-color complete'] + self.suffix

        return f'{bar}{self.colors["percentage complete"]}{percentage_str}\n' + self.colors['reset-color']

    def progress_failed(self, progress: float, total: float):
        percentage_str = ' Failed' if self.show_percentage else ''
        bar_length = self.width - len(percentage_str) - len(self.prefix) - len(self.suffix)

        if progress > total:
            bar_char = self.fill
        else:
            bar_char = self.empty

        bar = '\r' + self.colors['prefix-color failed'] + self.prefix + \
              self.colors['progress failed'] + bar_char * bar_length + \
              self.colors['suffix-color failed'] + self.suffix

        return f'{bar}{self.colors["percentage failed"]}{percentage_str}\n' + self.colors['reset-color']

    def __call__(self, progress: float, total: float, logger: '_log21.Logger' = None):
        if not logger:
            logger = self.logger

        logger.print(self.get_bar(progress, total), end='')

    def update(self, progress: float, total: float, logger: '_log21.Logger' = None):
        self(progress, total, logger)
